/* 
* Lavori in corso di 'openrtp.c', per il momento nominata 'rtplistn.c',
* completa di RX, DECODIFICA e PLYBACK.
***************************************************************************
*			   ***REAL-TIME***
***************************************************************************
									   */

#include <stdio.h>
#include <unistd.h>
#include <stddef.h>
#include <string.h>
#include <strings.h>	   // bcopy(), bzero()
#include <sys/soundcard.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>	   // signal()
#include <sys/socket.h>    // struct sockadrr 
#include <netinet/in.h>    // sockaddr_in 
#include <stdlib.h>	   // perror() 
#include "rtp.h"	   // RTP headers 
#include "typedef.h"
#include "basic_op.h"
#include "ld8a.h"


#define MAXLEN 450  // max num di byte associabili ad un pacchetto 
#define RECV 1

Word16 bad_lsf;        /* bad LSF indicator   */

/*
   This variable should be always set to zero unless transmission errors
   in LSP indices are detected.
   This variable is useful if the channel coding designer decides to
   perform error checking on these important parameters. If an error is
   detected on the  LSP indices, the corresponding flag is
   set to 1 signalling to the decoder to perform parameter substitution.
   (The flags should be set back to 0 for correct transmission).
*/


static char au_name[15] = "/dev/dsp";

int dsp_stereo = 0;    // al momento i codificatori disponibili sono tutti mono
int dsp_speed = 8000;  // PCM (defoult), G728, G729, ...
extern int au_fd;
char *audio_buff;
int dsp_speed, aubuf_size;

// list of received packet 
struct packlist {    
 char data[MAXLEN];	    // payload 
 struct packlist *next;     // pointer to next element in the list 
} *packinfo;
struct packlist *start, *last;	// pointer to first and last elements
struct packlist *prev=0;	// pointer to previus sent packet

// received packet to store in the list  
char packet[MAXLEN]; 

static int sockfd;  // input socket 
 
static int pcount = 0;
 

/*
 * Questa funzione fa l'operazione inversa della 'info_synt'.
 */

void info_extract( unsigned char payload[10], Word16 parm[PRM_SIZE+1])
{

 int i;
 unsigned short buff[10];

 for(i=0;i<10;i++)
  buff[i] = payload[i];

 parm[1]  = buff[0];
 parm[2]  = buff[1]+((buff[2]&192)<<2);
 parm[3]  = (buff[2]&63)+(buff[3]&192);
 parm[4]  = ((buff[3]&32)>>5)&1;
 parm[5]  = (buff[3]&31)+(buff[4]<<5);
 parm[6]  = ((buff[5]&240)>>4)&15;
 parm[7]  = (buff[5]&15)+(((buff[6]&224)>>1)&112);
 parm[8]  = buff[6]&31;
 parm[9]  = buff[7]+((buff[8]&248)<<5);
 parm[10] = (buff[8]&7)+(((buff[9]&128)>>4)&8);
 parm[11] = buff[9]&127;

}
 
 
 
/*
* Questa funzione inserisce in una lista i pacchetti in arrivo secondo
* l'ordine dettato dal 'sequence number'; jitter equalizzato, cos potranno 
* essere decodificati.
*/
static
void store(struct packlist *i, struct packlist **start, struct packlist **last, 
           int lseq)
{
 
 struct packlist *old, *p;
 rtp_hdr_t *hpi = (rtp_hdr_t *)(i->data);
 rtp_hdr_t *hpp;
 
 
 /*
  * se il pacchetto  in eccessivo ritardo, non va messo in lista.
  * 'lseq'  indica il sequenze number dell'ultimo pacchetto ascoltato.
  */
 if( ntohs(hpi->seq) < lseq ) 
  goto no_store;
  
 p = *start;
 hpp = (rtp_hdr_t *)(p->data);
 if(!*last) { // primo elemento della lista (prima volta) 
  i->next = NULL;
  *last = i;
  *start = i;
  return;
 }

 old = NULL;
 
 while(p) {
  if( ntohs(hpp->seq) < ntohs(hpi->seq) ) {
   old = p;
   p = p->next;
  }
  else {
   if(old) { // elemento mediano 
    old->next = i;
    i->next = p;
    return;
   }
   i->next = p; // primo elemento 
   *start = i;
   return;
  }
 } /* while */
 
 (*last)->next = i;  // ultimo elemento 
 i->next = NULL;
 *last = i;

 no_store: 

}  /* store */


/*
 * End function, print a short message for utent.
 */
static void done(int signo)
{
 
 fprintf(stderr,"Trasmission interrupted\n");
 fprintf(stderr,"%d packets received\n",pcount); 

 exit(0);
 
}


/*
 * main function 
 */
int main(void)
{

 struct sockaddr_in servaddr, cliaddr;
 int clilen;  
 int len; // # di byte per packet (header+payload); defoult: 20ms per packet 
 const int on = 1;
 int n;
 int padding = 0;  // # of padding-byte in a packet
 int pad = 0;	  // # of padding-byte not decoded
 rtp_hdr_t *hp;
 u_int16 lseq = 0;
 extern void set_snd(char *audio_name, int open_mod, int dsp_stereo);
 extern int g711(char type_of_conv[2], char *data_in, char *data_out,
		 int padding); 

  unsigned char payload[10];
  Word16 i;

  Word16  synth_buf[L_FRAME+M], *synth; /* Synthesis		       */
  Word16  parm[PRM_SIZE+1];		/* Synthesis parameters        */
  Word16  Az_dec[MP1*2];		/* Decoded Az for post-filter  */
  Word16  T2[2];			/* Pitch lag for 2 subframes   */

 /*
  * open/bind one network socket for RTP.
  */  
 bzero(&servaddr, sizeof(servaddr));

 servaddr.sin_family = AF_INET;
 servaddr.sin_port   = htons(5004);
 servaddr.sin_addr.s_addr = htonl(INADDR_ANY);

 sockfd = socket(AF_INET, SOCK_DGRAM, 0);
 if( sockfd < 0 ) {
  perror("socket");
  exit(1);
 }
  
 // Set SO_REUSEADDR socket option in case connection exist on this port 
 if( setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1 )
  perror("setsockopt (SO_REUSEADDR)");

 //  Set the socket receive buffer to 240 Kbytes
 n = 240 * 1024;
 if( setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &n, sizeof(n)) == -1 )
  perror("setsockopt (SO_RCVBUF)");

 if( bind(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0 ) {
  perror("bind");
  exit(1);
 }
 
/* 
 * Qua ci entrerebbe tutto un discorso su 'select' per la distinzione tra
 * pacchetti RTP ed RTCP e per il full-duplex [rtpdump.c,12/12] [Stevens,c.6];
 * per il momento tralasciamo questi aspetti 
 */
 start = last = NULL;

 // setup soundcard to enable playback.
 set_snd(au_name, O_WRONLY, dsp_stereo);								 
  
   /*-----------------------------------------------------------------*
   *	       Initialization of decoder                             * 
   * N.B. Le successive tre chiamate a funzioni non  necessario     *
   *      farle per ogni pacchetto, basta farle una volta per tutte. *
   *      Anzi  consigliato farle una volta per tutte in modo da non*
   *      aggiungere inutili ritardi all'algoritmo.		     *  			     *
   *-----------------------------------------------------------------*/
  

  for (i=0; i<M; i++) synth_buf[i] = 0;
   synth = synth_buf + M;

   bad_lsf = 0;		/* Initialize bad LSF indicator */
   Init_Decod_ld8a();
   Init_Post_Filter();
   Init_Post_Process();

 // Stop connection and print a short message
 signal(SIGINT, done);
 
 // main loop
 while(RECV) {
  
  fill_buffer:
  
  clilen = sizeof(cliaddr);
  
  
  len = recvfrom(sockfd, packet, MAXLEN, 
		 0,(struct sockaddr *)&cliaddr, &clilen);
	
  // Add the packet to the list of packets (equalyzing jitter from seq. num.)
  packinfo = (struct packlist *)malloc(sizeof(struct packlist));
  if(!packinfo) 
   fprintf(stderr,"Warning, no more memory for incoming packets\n");
  bcopy(packet , packinfo->data, len );
  store(packinfo, &start, &last, lseq); 
  
  pcount++;
   
  // jitter-buffer (20 packet) 
   /*
    * momentaneamente, la soluzione  quella di riempire il buffer
    * con un numero di pacchetti '20' sufficiente ad eliminare il
    * jitter e, contemporaneamente, a non ritardare troppo l'emissione
    * dei pacchetti verso il decodeificatore.
    * Il Jitter-Buffer, una volta riempito con i primi '20' pacchetti,
    * ne conterr sempre '20'.
    */
  if( pcount < 20 )
   goto fill_buffer;
   
  fprintf(stderr,"%d packet parsed from jitter-buffer\n",pcount);
  
  // update pointer 'hp' to packet to play
  hp = (rtp_hdr_t *)(start->data); 
  lseq = ntohs( hp->seq );
  
  // parse PT to adopt appropriate decoder.
  switch( hp->pt ) { 
  case 0: // PCMU
  case 8: // PCMA
     
    // parse padding field
    if( hp->p == 1 ) {
    padding = start->data[160-1];
    pad = start->data[160-1] * 2;
   }
    
    // begin appropriate decoding and depacking routines
    if( hp->pt == 8 )
     g711("al",&(start->data[12]), audio_buff, padding); 
    else 
     g711("ul",&(start->data[12]), audio_buff, padding);
     
    break;
  case 4: // G.723
  
    fprintf(stderr, 
	    "Sorry, G.723 coder is unavailable at moment\n");
    exit(0);
    
    break;
  case 15:  // G.728
  
    fprintf(stderr, 
	   "Sorry, G.728 coder is unavailable at moment\n");
    exit(0);
    
    break;
  case 18:  // G.729
  
    // N. B. padding field must to be always zero

   for(i=0;i<2;i++) {
    
   bcopy(&(start->data[12+i*10]), payload, 10);

   info_extract(payload, parm); // bit depacketization

   /* the hardware detects frame erasures by checking if all bits
      are set to zero
    */
   parm[0] = 0;		 /* No frame erasure */

   /* check pitch parity and put 1 in parm[4] if parity error */

   parm[4] = Check_Parity_Pitch(parm[3], parm[4]);

   Decod_ld8a(parm, synth, Az_dec, T2);

   Post_Filter(synth, Az_dec, T2);	 /* Post-filter */

   Post_Process(synth, L_FRAME);

   bcopy(synth, (Word16 *)audio_buff+i*80, 2*L_FRAME);
   }
    break;
  case 20: // MELP 
  
    fprintf(stderr, 
	    "Sorry, MELP coder is unavailable at moment\n");
    exit(0);
    
    break;
  }
  
  /*
   * Playback recorded sound. 
   * Every received payload is decoded in
   * linear PCM, 16 bits/sample ---> 20ms = 320 byte
   */
  if( write(au_fd, audio_buff, (320-pad) ) != (320-pad) ) {
   perror(au_name);
   exit(1);
  } 
 
  prev = start;
  start = start->next;
  free(prev);  
  /*
   * La lista, che verr scandita per la memorizzazione, conterr 
   * sempre solo '20' pacchetti e non di pi. In questo modo si diminuir
   * notevolmente il tempo di ricerca del punto di memorizzazione, 
   * arginando, si spera, il rischio di OVERFLOW.
   * 
   * Si libera la memoria, con free(), per evitare che una lunga 
   * conversazione venga meno per mancanza di memoria da allocare.
   */  

 } /* while */

return 0;

} /* main */

 


