/*
 ********************************************************************
 ********************************************************************
 * RTP-Speak.                                                       *
 * Principale routine di ricezione.                                 *
 * Versione 1.0 (Maggio 2001)                                       * 
 * Gestisce:                                                        *
 * 1. La ricezione dall'interfaccia di rete                         *
 * 2. La decodifica                                                 *
 * 3. Il playout                                                    *
 ********************************************************************
 ********************************************************************
 * Autore: Daniele Mari                                             *
 * e-mail: daniele.mari@tiscalinet.it                               *
 ********************************************************************
 ********************************************************************
                                                                    */


#include <stdio.h>
#include <unistd.h>
#include <stddef.h>
#include <string.h>
#include <strings.h>	   		// bcopy(), bzero()
#include <sys/soundcard.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>	   			// signal()
#include <sys/socket.h>    		// struct sockadrr 
#include <netinet/in.h>    		// sockaddr_in 
#include <stdlib.h>	   			// perror() 
#include "rtp.h"	   			// RTP headers 


#define MAXLEN 350  	 /*
						  * massimo num di byte associabili ad un     
                          * pacchetto RTP ( consigliato non superare 
                          * questa soglia) 
                          */

#define RECV 1


/*
 * Dichiarazioni globali per l'acqusizione audio
 *
 */

static char au_name[15] = "/dev/dsp";

int dsp_stereo = 0;    // mono
int dsp_speed = 8000;  // PCMA (defoult), PCMU, G729
extern int au_fd;
char *audio_buff;
int dsp_speed, aubuf_size;

/*
 * Lista dei pacchetti ricevuti (n.b. la lista  dinamica.)
 * I pacchetti in lista sono equalizzati dal jitter. 
 */ 
struct packlist {    
 char data[MAXLEN];	    	// RTP-header + payload 
 struct packlist *next;     // puntatore all'elemanto successivo
} *packinfo;
struct packlist *start, *last;	/*
  								 * puntatori al primo e all'ultimo 
                               	 * elemento della lista
								 */

struct packlist *prev=0;	// puntatore al pacchetto gi elaborato

// pacchetti ricevuti da equalizzare  
char packet[MAXLEN]; 

static int sockfd[2];  // input sockets  
 
static int pcount = 0;
 
 
 
/*
 * Questa funzione inserisce in una lista i pacchetti in arrivo 
 * secondo l'ordine dettato dal 'sequence number', jitter equalizzato, 
 * in modo che possano essere passati al decodificatore. 
 * La lista  aggiornata dinamicamente.
 */
static
void store(struct packlist *i, struct packlist **start, struct packlist **last, 
           int lseq)
{
 
 struct packlist *old, *p;
 rtp_hdr_t *hpi = (rtp_hdr_t *)(i->data);
 rtp_hdr_t *hpp;
 
 
 /*
  * se il pacchetto  in eccessivo ritardo, non va messo in lista.
  * 'lseq'  indica il sequenze number dell'ultimo pacchetto ascoltato.
  */
  
 if( ntohs(hpi->seq) < lseq )
  goto no_store;
  
 p = *start;
 hpp = (rtp_hdr_t *)(p->data);
 
 if(!*last) { // primo elemento della lista (prima volta) 
  i->next = NULL;
  *last = i;
  *start = i;
  return;
 }

 old = NULL;
 
 while(p) { 
  if( ntohs(hpp->seq) < ntohs(hpi->seq) ) {  
   old = p;
   p = p->next;
  }
  else { 
   if(old) { // elemento mediano 
    old->next = i;
    i->next = p;
    return;
   }
   i->next = p; // primo elemento 
   *start = i;
   return;
  }
 } /* while */
 
 (*last)->next = i;  // ultimo elemento 
 i->next = NULL;
 *last = i;

 no_store: 

}  /* store */


/*
 * Funzione di fine.
 * Stampa un breve messaggio di notifica.
 * E' innescata, in modo asincrono, dall'utente mediante 
 * il comando: ^C
 */
static void done(int signo)
{
 
 fprintf(stderr,"Trasmission interrupted\n");
 fprintf(stderr,"%d packets received\n",pcount);
 fprintf(stderr, "Bye :-)\n");
 fprintf(stderr, "\n");  

 exit(0);
 
}


/*
 * main function 
 */
int main(void)
{

 struct sockaddr_in servaddr, cliaddr;
 int clilen;  
 int len;           /*
       				 * # di byte per pacchetto (header+payload), 
 					 * defoult: 20ms per pacchetto. 
 					 */
 const int on = 1;
 int n,i;
 int padding = 0;  // # of padding-byte in a packet
 int pad = 0;	   // # of padding-byte not decoded 
 rtp_hdr_t *hp;
 u_int16 lseq = 0;
 extern void set_snd(char *audio_name, int open_mod, int dsp_stereo);
 extern int g711(char type_of_conv[2], char *data_in, char *data_out,
		 int padding); 


 /*
  * Si aprono due socket UDP non connesse, per i pacchetti RTP ed 
  * RTCP. 
  */  
 bzero(&servaddr, sizeof(servaddr));

 servaddr.sin_family = AF_INET;
 servaddr.sin_port   = htons(5004);
 servaddr.sin_addr.s_addr = htonl(INADDR_ANY);

  for(i=0;i<2;i++) {
  sockfd[i] = socket(AF_INET, SOCK_DGRAM, 0);
  if( sockfd[i] < 0 ) {
   perror("socket");
   exit(1);
  }
  servaddr.sin_port = htons( ntohs(servaddr.sin_port) + i );
  
 /*
  * Opzione SO_REUSEADDR. E' cautelativa nel caso sia stata gi aperta 
  * una connessione su questa porta. 
  */
 if( setsockopt(sockfd[i], SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1 )
  perror("setsockopt (SO_REUSEADDR)");

 //  Set the socket receive buffer to 240 Kbytes
 n = 240 * 1024;
 if( setsockopt(sockfd[i], SOL_SOCKET, SO_RCVBUF, &n, sizeof(n)) == -1 )
  perror("setsockopt (SO_RCVBUF)");

 if( bind(sockfd[i], (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0 ) {
  perror("bind");
  exit(1);
 }
 
} /* for */
 
/* 
 * Qua ci entrerebbe tutto un discorso su 'select' per la distinzione tra
 * pacchetti RTP ed RTCP e per il full-duplex [Stevens,c.6].
 */
 
 start = last = NULL;
 
 fprintf(stderr,"\n");
 fprintf(stderr,"------------------------------------------------\n");
 fprintf(stderr,"********************RTP-Speak*******************\n");
 fprintf(stderr,"**************Version 1.0, May 2001*************\n");
 fprintf(stderr,"\n");
 fprintf(stderr,">>>>>>>>>>>>Designed by: Daniele Mari<<<<<<<<<<<\n");
 fprintf(stderr,"------------------------------------------------\n");
 fprintf(stderr,"\n");
 fprintf(stderr, "\t escape character is : ^C\n");
 fprintf(stderr, "\n");

 
/*
 * Inizializzazione della scheda audio per il playback.
 */
 set_snd(au_name, O_WRONLY, dsp_stereo);								 

 signal(SIGINT, done);
 
 // main loop
 while(RECV) {
 
  fill_buffer:

  clilen = sizeof(cliaddr);
  
  len = recvfrom(sockfd[0], packet, MAXLEN, 
		 0,(struct sockaddr *)&cliaddr, &clilen);

	
 /*
  * Routine di immagazzinamento nella lista a jitter equalizzato.
  */	
  packinfo = (struct packlist *)malloc(sizeof(struct packlist));
  if(!packinfo) 
   fprintf(stderr,"Warning, no more memory for incoming packets\n");
  bcopy(packet , packinfo->data, len );
  store(packinfo, &start, &last, lseq); 
  
  pcount++;
   
  // jitter-buffer (20 pacchetti)
   /*
    * momentaneamente, la soluzione  quella di riempire il buffer
    * con un numero di pacchetti '20' sufficiente ad eliminare il
    * jitter e, contemporaneamente, a non ritardare troppo l'emissione
    * dei pacchetti verso il decodeificatore.
    * Il Jitter-Buffer, una volta riempito con i primi '20' pacchetti,
    * ne conterr sempre '20'.
    */
    
  if( pcount < 20 )
   goto fill_buffer;

  fprintf(stderr,"%d packet parsed from jitter-buffer\r",pcount);
  
  // aggiorna il puntatore'hp' al pacchetto da ascoltare.
  hp = (rtp_hdr_t *)(start->data); 
  lseq = ntohs( hp->seq );
  
  // Individua il decodificatore appropriato attraverso il PT
  switch( hp->pt ) { 
  case 0: // PCMU
  case 8: // PCMA

    // valuta il campo di padding
    if( hp->p == 1 ) {
    padding = start->data[160-1];
    pad = start->data[160-1] * 2;
   }
    
     /*
      * Innesca le appropriate routine di decodifica e 
      * depacchettizzazione
      */
    if( hp->pt == 8 )
     g711("al",&(start->data[12]), audio_buff, padding);
    else 
     g711("ul",&(start->data[12]), audio_buff, padding);
     
    break;
  case 18:  // G.729
  
    // parse padding field
    if( hp->p == 1 ) {
     padding = start->data[20-1];
     pad = start->data[20-1] * 16;
    }
    
    fprintf(stderr, 
	    "Sorry, G.729 coder is unavailable at moment\n");
    exit(0);
    
    break;
  }
  
  /*
   * Playback. 
   * Ogni pacchetto ricevuto  decodificato in
   * PCM lineare, 16 bits/campione ---> 20ms = 320 byte.
   * Attenzione al padding.
   */
  if( write(au_fd, audio_buff, (320-pad) ) != (320-pad) ) { 
   perror(au_name);
   exit(1);
  } 


  /*
   * La lista, che verr scandita per la memorizzazione, conterr 
   * sempre solo '20' pacchetti e non di pi. In questo modo si
   * diminuir notevolmente il tempo di ricerca del punto di 
   * memorizzazione, arginando, si spera, il rischio di OVERFLOW.
   * 
   * Si libera la memoria, con free(), per evitare che una lunga 
   * conversazione venga meno per mancanza di memoria da allocare.
   */  
   
  prev = start;
  start = start->next;
  free(prev);  

 } /* while */

return 0;

} /* main */

 


