/*
 ********************************************************************
 ********************************************************************
 * Questo file contiene la funzione 'set_snd()' che realizza routine* 
 * generali di setup, valide sia per la registrazione che per il    *
 * playback dell'audio.                                             *
 *                                                                  *
 * Versione 1.0 (Maggio 2001)                                       * 
 *                                                                  *
 * La 'vm_toall.c' la richiama cos:                                * 
 *   set_snd(au_name, O_RDONLY, dsp_stereo);                        *
 *                                                                  *
 * la 'oplnall.c'  la richiama cos:                                *
 *   set_snd(au_name, O_WRONLY, dsp_stereo);                        *
 ********************************************************************
 ********************************************************************
 * Autore: Daniele Mari                                             *
 * e-mail: daniele.mari@tiscalinet.it                               *
 ********************************************************************
 ********************************************************************
                                                                    */

#include <stdio.h>
#include <stdlib.h>          // perror()
#include <sys/ioctl.h>       // ioctl()
#include <fcntl.h>
#include <sys/soundcard.h>


int samplesize = 16;  		// (defoult) PCM-lineare 16 bit/sample
extern int dsp_speed, aubuf_size;
extern char *audio_buff;
int au_fd;
int profile = APF_NORMAL;


void set_snd(char *audio_name, int open_mod, int dsp_stereo) 
{
 int app;	
 
 if( (au_fd = open(audio_name, open_mod, 0)) == -1 ) {
  perror(audio_name);
  exit(1);
 }

 // set-up del numeri di bit per campione.
 app = samplesize;
 ioctl(au_fd, SNDCTL_DSP_SAMPLESIZE, &samplesize);
 if( app != samplesize ) {
  fprintf(stderr, "Unable to set the samplesize (%d bit)\n",app);

  exit(1);
 }

 // set-up del profilo audio
 ioctl(au_fd,SNDCTL_DSP_PROFILE, &profile);

 // set-up del numero di canali  2stereo/1mono
 if( ioctl(au_fd, SNDCTL_DSP_STEREO, &dsp_stereo) == -1 ) {
  fprintf(stderr, "Unable to set stereo/mono\n");
  exit(1);
 }

 // set-up dei parametri di campionamento
 if( ioctl(au_fd, SNDCTL_DSP_SPEED, &dsp_speed) == -1 ) {
  fprintf(stderr, "Unable to set audio speed (sampling frequency)\n");
  exit(1);
 }

 // riscontro sulla taglia del buffer audio
 ioctl(au_fd, SNDCTL_DSP_GETBLKSIZE, &aubuf_size);
 if( aubuf_size < 1 ) {
  perror("SNDCTL_DSP_GETBLKSIZE");
  exit(1);
 }

 /*
  * Allocazione dinamica della memoria per il buffer audio.
  *
  * Dopo che registro vado a leggere da 'au_fd' 320 byte (20ms di PCM)
  * e li pongo in 'audio_buf'; quindi, se 'aubuf_size' di defoult  
  * minore dovr allocare pi memoria, altrimenti vanno bene 
  * aubuf_size-byte.
  * Per il playback vale la stessa cosa.
  */
 if ( aubuf_size < 320 ) 
  aubuf_size = 320;
  if( (audio_buff = malloc(aubuf_size)) == NULL ) { 
   fprintf(stderr, "Unable to allocate memory for I/O audio 
           buffer\n");
   exit(1);
  }

 // messaggio per l'utente 
 fprintf(stderr, "Speed %d Hz %d bps", dsp_speed, samplesize);
 if( dsp_stereo )
  fprintf(stderr, "(stereo)\n");
 else
  fprintf(stderr, "(mono)\n");
/* 
 * N.B. quest'ultimi messaggi li ho scritti su 'stderr' anche se non
 * sono messaggi d'errore perch 'stdout' potrebbe essere occupato,     
 * per defoult, come canale d'uscita dei dati audio.     
 */

} /* set_snd */
